@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """The prompt_toolkit based xonsh shell."""
 import os
 import re
@@ -6,37 +5,36 @@ import sys
 from functools import wraps
 from types import MethodType
 
-from xonsh.built_ins import XSH
-from xonsh.events import events
-from xonsh.base_shell import BaseShell
-from xonsh.ptk_shell.formatter import PTKPromptFormatter
-from xonsh.shell import transform_command
-from xonsh.tools import print_exception, print_warning, carriage_return
-from xonsh.platform import HAS_PYGMENTS, ON_WINDOWS, ON_POSIX
-from xonsh.style_tools import partial_color_tokenize, _TokenType, DEFAULT_STYLE_DICT
-from xonsh.lazyimps import pygments, pyghooks, winutils
-from xonsh.pygments_cache import get_all_styles
-from xonsh.ptk_shell.history import PromptToolkitHistory, _cust_history_matches
-from xonsh.ptk_shell.completer import PromptToolkitCompleter
-from xonsh.ptk_shell.key_bindings import load_xonsh_bindings
-
 from prompt_toolkit import ANSI
 from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
-from prompt_toolkit.lexers import PygmentsLexer
 from prompt_toolkit.enums import EditingMode
+from prompt_toolkit.formatted_text import PygmentsTokens, to_formatted_text
+from prompt_toolkit.history import ThreadedHistory
 from prompt_toolkit.key_binding.bindings.emacs import (
     load_emacs_shift_selection_bindings,
 )
-from prompt_toolkit.key_binding.key_bindings import merge_key_bindings
 from prompt_toolkit.key_binding.bindings.named_commands import get_by_name
-from prompt_toolkit.history import ThreadedHistory
-from prompt_toolkit.shortcuts import print_formatted_text as ptk_print
+from prompt_toolkit.key_binding.key_bindings import merge_key_bindings
 from prompt_toolkit.shortcuts import CompleteStyle
+from prompt_toolkit.shortcuts import print_formatted_text as ptk_print
 from prompt_toolkit.shortcuts.prompt import PromptSession
-from prompt_toolkit.formatted_text import PygmentsTokens, to_formatted_text
-from prompt_toolkit.styles import merge_styles, Style
+from prompt_toolkit.styles import Style, merge_styles
 from prompt_toolkit.styles.pygments import pygments_token_to_classname
 
+from xonsh.base_shell import BaseShell
+from xonsh.built_ins import XSH
+from xonsh.events import events
+from xonsh.lazyimps import pyghooks, pygments, winutils
+from xonsh.platform import HAS_PYGMENTS, ON_POSIX, ON_WINDOWS
+from xonsh.ptk_shell.completer import PromptToolkitCompleter
+from xonsh.ptk_shell.formatter import PTKPromptFormatter
+from xonsh.ptk_shell.history import PromptToolkitHistory, _cust_history_matches
+from xonsh.ptk_shell.key_bindings import load_xonsh_bindings
+from xonsh.pygments_cache import get_all_styles
+from xonsh.shell import transform_command
+from xonsh.style_tools import DEFAULT_STYLE_DICT, _TokenType, partial_color_tokenize
+from xonsh.tools import carriage_return, print_exception, print_warning
+
 try:
     from prompt_toolkit.clipboard import DummyClipboard
     from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
@@ -45,7 +43,13 @@ try:
 except ImportError:
     HAVE_SYS_CLIPBOARD = False
 
-ANSI_OSC_PATTERN = re.compile("\x1b].*?\007")
+try:
+    from prompt_toolkit.cursor_shapes import ModalCursorShapeConfig
+
+    HAVE_CURSOR_SHAPE = True
+except ImportError:
+    HAVE_CURSOR_SHAPE = False
+
 CAPITAL_PATTERN = re.compile(r"([a-z])([A-Z])")
 Token = _TokenType()
 
@@ -86,19 +90,6 @@ def tokenize_ansi(tokens):
     return ansi_tokens
 
 
-def remove_ansi_osc(prompt):
-    """Removes the ANSI OSC escape codes - ``prompt_toolkit`` does not support them.
-    Some terminal emulators - like iTerm2 - uses them for various things.
-
-    See: https://www.iterm2.com/documentation-escape-codes.html
-    """
-
-    osc_tokens = ANSI_OSC_PATTERN.findall(prompt)
-    prompt = ANSI_OSC_PATTERN.sub("", prompt)
-
-    return prompt, osc_tokens
-
-
 def _pygments_token_to_classname(token):
     """Converts pygments Tokens, token names (strings) to PTK style names."""
     if token and isinstance(token, str):
@@ -197,6 +188,7 @@ class PromptToolkitShell(BaseShell):
             winutils.enable_virtual_terminal_processing()
         self._first_prompt = True
         self.history = ThreadedHistory(PromptToolkitHistory())
+        self.push = self._push
 
         ptk_args.setdefault("history", self.history)
         if not XSH.env.get("XONSH_COPY_ON_DELETE", False):
@@ -205,9 +197,10 @@ class PromptToolkitShell(BaseShell):
             ptk_args.setdefault("clipboard", PyperclipClipboard())
         self.prompter: PromptSession = PromptSession(**ptk_args)
 
-        self.prompt_formatter = PTKPromptFormatter(self.prompter)
+        self.prompt_formatter = PTKPromptFormatter(self)
         self.pt_completer = PromptToolkitCompleter(self.completer, self.ctx, self)
-        self.key_bindings = load_xonsh_bindings()
+        ptk_bindings = self.prompter.app.key_bindings
+        self.key_bindings = load_xonsh_bindings(ptk_bindings)
         self._overrides_deprecation_warning_shown = False
 
         # Store original `_history_matches` in case we need to restore it
@@ -225,6 +218,64 @@ class PromptToolkitShell(BaseShell):
             [self.key_bindings, load_emacs_shift_selection_bindings()]
         )
 
+    def get_lazy_ptk_kwargs(self):
+        """These are non-essential attributes for the PTK shell to start.
+        Lazy loading these later would save some startup time.
+        """
+        if not XSH.env.get("COLOR_INPUT"):
+            return
+
+        if HAS_PYGMENTS:
+            # these imports slowdown a little
+            from prompt_toolkit.lexers import PygmentsLexer
+
+            yield "lexer", PygmentsLexer(pyghooks.XonshLexer)
+
+        events.on_timingprobe.fire(name="on_pre_prompt_style")
+        yield "style", self.get_prompt_style()
+        events.on_timingprobe.fire(name="on_post_prompt_style")
+
+    def get_prompt_style(self):
+        env = XSH.env
+
+        style_overrides_env = env.get("PTK_STYLE_OVERRIDES", {}).copy()
+        if (
+            len(style_overrides_env) > 0
+            and not self._overrides_deprecation_warning_shown
+        ):
+            print_warning(
+                "$PTK_STYLE_OVERRIDES is deprecated, use $XONSH_STYLE_OVERRIDES instead!"
+            )
+            self._overrides_deprecation_warning_shown = True
+        style_overrides_env.update(env.get("XONSH_STYLE_OVERRIDES", {}))
+
+        if HAS_PYGMENTS:
+            style = _style_from_pygments_cls(pyghooks.xonsh_style_proxy(self.styler))
+            if len(self.styler.non_pygments_rules) > 0:
+                try:
+                    style = merge_styles(
+                        [
+                            style,
+                            _style_from_pygments_dict(self.styler.non_pygments_rules),
+                        ]
+                    )
+                except (AttributeError, TypeError, ValueError) as style_exception:
+                    print_warning(
+                        f"Error applying style override!\n{style_exception}\n"
+                    )
+
+        else:
+            style = _style_from_pygments_dict(DEFAULT_STYLE_DICT)
+
+        if len(style_overrides_env) > 0:
+            try:
+                style = merge_styles(
+                    [style, _style_from_pygments_dict(style_overrides_env)]
+                )
+            except (AttributeError, TypeError, ValueError) as style_exception:
+                print_warning(f"Error applying style override!\n{style_exception}\n")
+        return style
+
     def singleline(
         self, auto_suggest=None, enable_history_search=True, multiline=True, **kwargs
     ):
@@ -251,6 +302,10 @@ class PromptToolkitShell(BaseShell):
         completer = None if completions_display == "none" else self.pt_completer
 
         events.on_timingprobe.fire(name="on_pre_prompt_tokenize")
+
+        # clear prompt level cache
+        env["PROMPT_FIELDS"].reset()
+
         get_bottom_toolbar_tokens = self.bottom_toolbar_tokens
         if env.get("UPDATE_PROMPT_ON_KEYPRESS"):
             get_prompt_tokens = self.prompt_tokens
@@ -277,6 +332,11 @@ class PromptToolkitShell(BaseShell):
         ):
             self.prompter.default_buffer._history_matches = self._history_matches_orig
 
+        menu_rows = env.get("COMPLETIONS_MENU_ROWS", None)
+        if menu_rows:
+            # https://github.com/xonsh/xonsh/pull/4477#pullrequestreview-767982976
+            menu_rows += 1
+
         prompt_args = {
             "mouse_support": mouse_support,
             "auto_suggest": auto_suggest,
@@ -288,7 +348,7 @@ class PromptToolkitShell(BaseShell):
             "editing_mode": editing_mode,
             "prompt_continuation": self.continuation_tokens,
             "enable_history_search": enable_history_search,
-            "reserve_space_for_menu": 0,
+            "reserve_space_for_menu": menu_rows,
             "key_bindings": self.key_bindings,
             "complete_style": complete_style,
             "complete_while_typing": complete_while_typing,
@@ -296,60 +356,15 @@ class PromptToolkitShell(BaseShell):
             "refresh_interval": refresh_interval,
             "complete_in_thread": complete_in_thread,
         }
-
-        if env.get("COLOR_INPUT"):
-            events.on_timingprobe.fire(name="on_pre_prompt_style")
-            style_overrides_env = env.get("PTK_STYLE_OVERRIDES", {}).copy()
-            if (
-                len(style_overrides_env) > 0
-                and not self._overrides_deprecation_warning_shown
-            ):
-                print_warning(
-                    "$PTK_STYLE_OVERRIDES is deprecated, use $XONSH_STYLE_OVERRIDES instead!"
-                )
-                self._overrides_deprecation_warning_shown = True
-            style_overrides_env.update(env.get("XONSH_STYLE_OVERRIDES", {}))
-
-            if HAS_PYGMENTS:
-                prompt_args["lexer"] = PygmentsLexer(pyghooks.XonshLexer)
-                style = _style_from_pygments_cls(
-                    pyghooks.xonsh_style_proxy(self.styler)
-                )
-                if len(self.styler.non_pygments_rules) > 0:
-                    try:
-                        style = merge_styles(
-                            [
-                                style,
-                                _style_from_pygments_dict(
-                                    self.styler.non_pygments_rules
-                                ),
-                            ]
-                        )
-                    except (AttributeError, TypeError, ValueError) as style_exception:
-                        print_warning(
-                            f"Error applying style override!\n{style_exception}\n"
-                        )
-
-            else:
-                style = _style_from_pygments_dict(DEFAULT_STYLE_DICT)
-
-            if len(style_overrides_env) > 0:
-                try:
-                    style = merge_styles(
-                        [style, _style_from_pygments_dict(style_overrides_env)]
-                    )
-                except (AttributeError, TypeError, ValueError) as style_exception:
-                    print_warning(
-                        f"Error applying style override!\n{style_exception}\n"
-                    )
-
-            prompt_args["style"] = style
-            events.on_timingprobe.fire(name="on_post_prompt_style")
-
         if env["ENABLE_ASYNC_PROMPT"]:
             # once the prompt is done, update it in background as each future is completed
             prompt_args["pre_run"] = self.prompt_formatter.start_update
+        else:
+            for attr, val in self.get_lazy_ptk_kwargs():
+                prompt_args[attr] = val
 
+        if editing_mode == EditingMode.VI and HAVE_CURSOR_SHAPE:
+            prompt_args["cursor"] = ModalCursorShapeConfig()
         events.on_pre_prompt.fire()
         line = self.prompter.prompt(**prompt_args)
         events.on_post_prompt.fire()
@@ -366,7 +381,9 @@ class PromptToolkitShell(BaseShell):
         src = "".join(self.buffer)
         src = transform_command(src)
         try:
-            code = self.execer.compile(src, mode="single", glbs=self.ctx, locs=None)
+            code = self.execer.compile(
+                src, mode="single", glbs=self.ctx, locs=None, compile_empty_tree=False
+            )
             self.reset_buffer()
         except Exception:  # pylint: disable=broad-except
             self.reset_buffer()
@@ -379,7 +396,6 @@ class PromptToolkitShell(BaseShell):
         if intro:
             print(intro)
         auto_suggest = AutoSuggestFromHistory()
-        self.push = self._push
         while not XSH.exit:
             try:
                 line = self.singleline(auto_suggest=auto_suggest)
@@ -411,16 +427,6 @@ class PromptToolkitShell(BaseShell):
         except Exception:  # pylint: disable=broad-except
             print_exception()
 
-        p, osc_tokens = remove_ansi_osc(p)
-
-        if kwargs.get("handle_osc_tokens"):
-            # handle OSC tokens
-            for osc in osc_tokens:
-                if osc[2:4] == "0;":
-                    env["TITLE"] = osc[4:-1]
-                else:
-                    print(osc, file=sys.__stdout__, flush=True)
-
         toks = partial_color_tokenize(p)
 
         return tokenize_ansi(PygmentsTokens(toks))
@@ -431,7 +437,7 @@ class PromptToolkitShell(BaseShell):
             carriage_return()
             self._first_prompt = False
 
-        tokens = self._get_prompt_tokens("PROMPT", "message", handle_osc_tokens=True)
+        tokens = self._get_prompt_tokens("PROMPT", "message")
         self.settitle()
         return tokens
 
@@ -457,7 +463,7 @@ class PromptToolkitShell(BaseShell):
         """Displays dots in multiline prompt"""
         if is_soft_wrap:
             return ""
-        width = width - 1
+        width -= 1
         dots = XSH.env.get("MULTILINE_PROMPT")
         dots = dots() if callable(dots) else dots
         if not dots:
@@ -529,7 +535,11 @@ class PromptToolkitShell(BaseShell):
                 ]
             )
         ptk_print(
-            tokens, style=proxy_style, end=end, include_default_pygments_style=False
+            tokens,
+            style=proxy_style,
+            end=end,
+            include_default_pygments_style=False,
+            **kwargs,
         )
 
     def color_style_names(self):
